/*
    Trabalho Prtico IV - IPDI
    -------------------------
    Anderson de Rezende Rocha <anderson.rocha@ic.unicamp.br>
    Leyza E. Baldo Dorini <leyzad@yahoo.com.br>
    Comparao de duas metologias para REALCE (Sharpening) de imagens

    OBSERVAO
    ----------
    Imagem[i,j] = *(Imagem + i * nc + j)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "imagem.h"
#include "morfo.h"
#include "utils.h"

// Constante para alguma referncia de nmeros muito grandes
#define INFINITO 99999999
// Referncia ao ponteiro /dev/null
#define null 0

static int TAM_JANELA = 3;
static int MAX_NIVEL_CINZA;
static int NRO_ESCALAS;

// Abordagem 1 para realce
void sharpening(int *imagem, int *imagemSaida, int nl, int nc, int B[TAM_JANELA][TAM_JANELA], int it, int *contador);
// Abordagem 2 para realce
void sharpening2(int *imagem, int *imagemSaida, int nl, int nc, int B[TAM_JANELA][TAM_JANELA], int it, int *contador);
// Zera uma imagem de entrada
void zerarImagem(int *imagem, int nl, int nc);
// Imprime o conteudo de uma imagem PGM
void imprimirImagem(int *imagem, int nl, int nc);
// Copia uma imagem pixel a pixel e nao apenas troca as referencias
void copiarImagem(int *fonte, int *destino,  int nl, int nc);
// Erosao e/ou Dilatacao com um elemento estruturante planar ou nao-planar
int erosaoDilatacaoMMDE(int *imagem, int *imagemSaida, int nl, int nc, int B[TAM_JANELA][TAM_JANELA],
                         int op, int contador);
// Dada uma configuracao de escala e tipo cria o elemento estruturante a partir
// de equacoes reconhecidas pelo sistema
void criarElementoEstruturante(int B[TAM_JANELA][TAM_JANELA], int tipoB, float escala);
// Conta os minimos de uma imagem processada pelo programa "reconstroi-vincent"
void contagemMinimos(int *imagem, int nl, int nc);
// Le a lista de escalas passadas por arquivo para aplicacao do filtro alternado
// sequencial
void configurarParametros(char *nomeArquivo,  float *listaEscalas);

int main (int argc, char *argv[])
{
    // tipoB eh o tipo do elemento estruturante, iteracoes diz o quanto
    // uma operacao deve ser executada
    int OK, nc, nl, mn, i, j, contador, iteracoes, tipoB;
    int nroIteracoesRealizadas;
    int *imagemEntrada, *imagemSaida, *imagemAnterior;
    // Escala considerada para criao dos elementos estruturantes
    float escala;
    float *listaEscalas;
    int B[TAM_JANELA][TAM_JANELA];

    printf("\nComparacao de 2 metodologias para REALCE de imagens\n");

    if (argc == 7)
    {
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);
        MAX_NIVEL_CINZA = mn;
        aloca_memo (&imagemSaida, nl, nc);
        aloca_memo (&imagemAnterior, nl, nc);
        info_imagem (argv[1], nl, nc, mn);

        tipoB = atoi(argv[5]);
        escala = atof(argv[6]);

        // Cria o elemento estruturante baseado na escala e no tipo
        // de equacao escolhido
        criarElementoEstruturante(B, tipoB, escala);
        switch(atoi(argv[3]))
        {
            case 0:
                // Realce com a abordagem 2
                iteracoes = atoi(argv[4]);
                if(iteracoes == -1)
                    iteracoes = INFINITO;

                nroIteracoesRealizadas = 0;
                sharpening2(imagemEntrada, imagemSaida, nl, nc, B, iteracoes, &nroIteracoesRealizadas);
                printf("Foram executadas %d iteracoes\n\n", nroIteracoesRealizadas);
                break;
            case 1:
                // Realce com a abordagem 1
                iteracoes = atoi(argv[4]);

                if(iteracoes == -1)
                    iteracoes = INFINITO;

                nroIteracoesRealizadas = 0;
                sharpening(imagemEntrada, imagemSaida, nl, nc, B, iteracoes, &nroIteracoesRealizadas);
                printf("Foram executadas %d iteracoes\n\n", nroIteracoesRealizadas);
                break;
            case 2:
                // Erosao escolhida feita ate idempotencia ou com n passos
                contador = 0;
                iteracoes = atoi(argv[4]);

                if(iteracoes == -1)
                    iteracoes = INFINITO;

                copiarImagem(imagemEntrada, imagemAnterior, nl, nc);
                while(!erosaoDilatacaoMMDE(imagemAnterior, imagemSaida, nl, nc, B, 0, contador++)
                      && contador < iteracoes)
                {
                    printf("Contador = %d\n", contador);
                    copiarImagem(imagemSaida, imagemAnterior, nl, nc);
                }
                break;
            case 3:
                // Dilatacao escolhida feita ate idempotencia ou com n passos
                contador = 0;
                iteracoes = atoi(argv[4]);

                if(iteracoes == -1)
                    iteracoes = INFINITO;

                copiarImagem(imagemEntrada, imagemAnterior, nl, nc);
                while(!erosaoDilatacaoMMDE(imagemAnterior, imagemSaida, nl, nc, B, 1, contador++)
                      && contador < iteracoes)
                {
                    printf("Contador = %d\n", contador);
                    copiarImagem(imagemSaida, imagemAnterior, nl, nc);
                }
                break;
            default:
                printf("Opcao nao reconhecida\n");
        }

        if(imagemEntrada != null)
        {
            grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
        }
        else
        {
            printf("No foi possivel gravar o resultado\n");
        }

        desaloca_memo(&imagemSaida);
        desaloca_memo(&imagemAnterior);
    }
    else if(argc == 2)
    {
        // Dada uma imagem "diferenca" produzida pelo programa "Reconstroi-Vincent"
        // conta o nmero de maximos regionais na mesma
        // Observacao: a imagem "diferenca" gerada e PBM (preto e branco)
        // entao deve ser salva como PGM no gimp primeiro para
        // poder ser processada. Isto porque o procedimento
        // de leitura para PGM do "imagem.h" nao esta funcionando
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);
        contagemMinimos(imagemEntrada, nl, nc);
    }
    else if(argc == 6)
    {
        // A partir da lista de escalas presentes no arquivo lido,
        // aplica o filtro alternado sequencial (...(d(e(d(e))))...)
        printf("FILTRO ALTERNADO SEQUENCIAL (...(d(e(d(e)))...))\n");
        OK = le_imagem_pgm (argv[1], &imagemEntrada, &nl, &nc, &mn);
        info_imagem (argv[1], nl, nc, mn);
        MAX_NIVEL_CINZA = mn;
        aloca_memo (&imagemSaida, nl, nc);
        aloca_memo (&imagemAnterior, nl, nc);

        NRO_ESCALAS = atoi(argv[3]);
        listaEscalas = calloc(NRO_ESCALAS, sizeof(float));
        tipoB = atoi(argv[4]);
        configurarParametros(argv[5], listaEscalas);

        printf("\nEscalas escolhidas para o FAS\n");

        for(i = 0; i < NRO_ESCALAS; i++)
            printf("%f\n", *(listaEscalas + i));

        copiarImagem(imagemEntrada, imagemAnterior, nl, nc);
        for(i = 0; i < NRO_ESCALAS; i++)
        {
            escala = *(listaEscalas + i);

            printf("Escala escolhida = %f\n", escala);
            criarElementoEstruturante(B, tipoB, escala);
            erosaoDilatacaoMMDE(imagemAnterior, imagemSaida, nl, nc, B, 0, contador++);
            copiarImagem(imagemSaida, imagemAnterior, nl, nc);
            zerarImagem(imagemSaida, nl, nc);
            erosaoDilatacaoMMDE(imagemAnterior, imagemSaida, nl, nc, B, 1, contador++);
        }

        if(imagemEntrada != null)
        {
            grava_imagem_pgm (imagemSaida, argv[2], nl, nc, mn);
            printf("Imagem \"%s\" gravada com sucesso!\n", argv[2]);
        }

        desaloca_memo(&imagemSaida);
        desaloca_memo(&imagemAnterior);
    }
    else
    {
        printf("\n1) mmde entrada.pgm saida.pgm <operacao> <iteracoes (-1 = idempotencia)>\n");
        printf("   <tipo elem (1 = paraboloide, 2 = cone, 3 = plano)> <escala>\n");
        printf("   (faz o realce com elementos estruturantes diferentes)\n");

        printf("\n2) mmde entrada.pgm\n");
        printf("   calcula os minimos de uma imagem processada pelo Reconstroi-Vincent\n");

        printf("\n3) mmde entrada.pgm saida.pgm <nroEscalas> <tipo elemento> <arquivo com escalas>\n");
        printf("   (aplica o filtro alternado sequencial a partir de diferentes escalas)\n");
    }

    printf("\n");

    return(0);
}

// Cria o elemento estruturante a partir da equacao reconhecida pelo sistema
void criarElementoEstruturante(int B[TAM_JANELA][TAM_JANELA], int tipoB, float escala)
{
    int i, j;
	int valor; 

	valor = -1;
    switch(tipoB)
    {
        case 1:
            // Paraboloide de equacao = g(x,y) = -(x^2 + y^2)
            for(i = 0; i < TAM_JANELA; i++)
                for(j = 0; j < TAM_JANELA; j++)
                    B[i][j] = (-1.0 * (pow(i - TAM_JANELA/2, 2)/escala + pow(j - TAM_JANELA/2, 2)/escala));
        break;
        case 2:
             // Cone de equacao = g(x,y) = ver site http://mathworld.wolfram.com/Cone.html
            valor = (int)(valor / escala);
            for(i = 0; i < TAM_JANELA; i++)
                for(j = 0; j < TAM_JANELA; j++)
                    B[i][j] = valor;

            B[1][1] = 0;                    
            break;
        case 3:
            // Elemento plano
            for(i = 0; i < TAM_JANELA; i++)
            {
                for(j = 0; j < TAM_JANELA; j++)
                {
                    B[i][j] = -(int)escala;
                }
            }
            break;
        default:
            break;
    }

    // Apenas imprimindo o elemento estruturante gerado
    printf("Elemento estruturante\n");
    for(i = 0; i < TAM_JANELA; i++)
    {
        for(j = 0; j < TAM_JANELA; j++)
        {
            printf("%d ", B[i][j]);
        }
        printf("\n");
    }
}

// Conta os minimos de uma imagem "diferenca" produzida pelo
// programa "reconstroi-vincent"
void contagemMinimos(int *imagem, int nl, int nc)
{
    int i, minimos;

    minimos = 0;

    for(i = 0; i < nl * nc; i++)
    {
        if(*(imagem + i) == 255)
            minimos++;
    }

    printf("Minimos regionais = %d\n", minimos);
}

// Imprime o conteudo de uma imagem PGM
void imprimirImagem(int *imagem, int nl, int nc)
{
    int i, j;

    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nc; j++)
            printf("%d ", *(imagem + i * nc + j));

        printf("\n");
    }
}

// Abordagem 2: faz a erosao ate idempotencia, dilatacao ate idempotencia e
// depois faz a comparacao com a imagem original para saber o realce
// OBS: esta abordagem esta sendo feita porque do outro jeito nao atinge
// a idempotencia. Executei ate 100.000 passos da outra forma e nada.
void sharpening2(int *imagem, int *imagemSaida, int nl, int nc, int B[TAM_JANELA][TAM_JANELA], int it, int *cont)
{
    int *imagemErodida;
    int *imagemDilatada;
    int *imagemAnterior;
    int i, j;
    int dMenosF, fMenosE;
    int idempotencia;
    int passos;
    int contador;

    aloca_memo (&imagemErodida, nl, nc);
    aloca_memo (&imagemDilatada, nl, nc);
    aloca_memo (&imagemAnterior, nl, nc);

    copiarImagem(imagem, imagemAnterior, nl, nc);

    // Faz a erosao ate idempotencia ou ate n passos
    contador = 0;
    while((!erosaoDilatacaoMMDE(imagemAnterior, imagemSaida, nl, nc, B, 0, ++contador)) && contador < it)
        copiarImagem(imagemSaida, imagemAnterior, nl, nc);

    copiarImagem(imagemAnterior, imagemErodida, nl, nc);
    copiarImagem(imagem, imagemAnterior, nl, nc);

    // Faz a dilatacao ate idempotencia ou ate n passos
    *cont = contador;
    contador = 0;
    while((!erosaoDilatacaoMMDE(imagemAnterior, imagemSaida, nl, nc, B, 1, ++contador)) && contador < it)
        copiarImagem(imagemSaida, imagemAnterior, nl, nc);

    // Testando o numero de iteracoes realizado
    if(contador > *cont)
        *cont = contador;

    copiarImagem(imagemAnterior, imagemDilatada, nl, nc);

    for(i = 0; i < nl; i++)
    {
        for(j =0; j < nc; j++)
        {
            // Fazendo o realce a partir da definicao proposta em
            // no artigo de J.G.M Schavemaker
            dMenosF = *(imagemDilatada + i * nc + j) - *(imagem + i * nc + j);
            fMenosE = *(imagem + i * nc + j) - *(imagemErodida + i * nc + j);

            if(dMenosF < fMenosE)
            {
                *(imagemSaida + i * nc + j) = *(imagemDilatada + i * nc + j);
            }
            else if(dMenosF > fMenosE)
            {
                *(imagemSaida + i * nc + j) = *(imagemErodida + i * nc + j);
            }
            else
            {
                *(imagemSaida + i * nc + j) = *(imagem + i * nc + j);
            }
        }
    }

    desaloca_memo(&imagemDilatada);
    desaloca_memo(&imagemErodida);
    desaloca_memo(&imagemAnterior);
}

// Abordagem 1: a cada passo compara com a imagem original e seleciona um elemento
// da dilatacao ou da erosao
// Nao sei se esta abordagem seria correta pois nao se atinge a idempotencia
void sharpening(int *imagem, int *imagemSaida, int nl, int nc, int B[TAM_JANELA][TAM_JANELA], int it, int *cont)
{
    int *imagemErodida;
    int *imagemDilatada;
    int *imagemAnterior;
    int i, j;
    int dMenosF, fMenosE;
    int idempotencia;
    int passos;
    int contador;

    aloca_memo (&imagemErodida, nl, nc);
    aloca_memo (&imagemDilatada, nl, nc);
    aloca_memo (&imagemAnterior, nl, nc);

    // Numero de passos ate atingir a idempotencia
    passos = 0;
    // Nao atingiu idempotencia
    idempotencia = 0;
    // Contador de diferencias
    contador = 0;

    copiarImagem(imagem, imagemAnterior, nl, nc);

    for(i = 0; i < nl; i++)
    {
        *(imagemAnterior + i * nc) = 0;
        *(imagemAnterior + i * nc + (nl - 1)) = 0;
    }

    for(j = 0; j < nc; j++)
    {
        *(imagemAnterior + j) = 0;
        *(imagemAnterior + (nl - 1) * nc + j) = 0;
    }

    while(idempotencia == 0 && passos < it)
    {
        // Um passo adicional pois a idempotencia ainda nao foi
        // atingida
        passos++;

        // Apenas para dar um feedback de execucacao ao usuario
        if(passos % 100 == 0)
            printf("Passos realizados = %d\n", passos);

        // Preparando imagens para processamento
        zerarImagem(imagemErodida, nl, nc);
        zerarImagem(imagemDilatada, nl, nc);
        zerarImagem(imagemSaida, nl, nc);

        // Faz uma erosao e uma dilatacao
        erosaoDilatacaoMMDE(imagemAnterior, imagemErodida, nl, nc, B, 0, 0);
        erosaoDilatacaoMMDE(imagemAnterior, imagemDilatada, nl, nc, B, 1, 0);

        for(i = 0; i < nl; i++)
        {
            for(j =0; j < nc; j++)
            {
                // Faz o realce segundo o artigo de Schavemaker
                dMenosF = *(imagemDilatada + i * nc + j) - *(imagem + i * nc + j);
                fMenosE = *(imagem + i * nc + j) - *(imagemErodida + i * nc + j);

                if(dMenosF < fMenosE)
                {
                    *(imagemSaida + i * nc + j) = *(imagemDilatada + i * nc + j);
                }
                else if(dMenosF > fMenosE)
                {
                    *(imagemSaida + i * nc + j) = *(imagemErodida + i * nc + j);
                }
                else
                {
                    *(imagemSaida + i * nc + j) = *(imagem + i * nc + j);
                }
            }
        }

        // Atribuindo o numero de iteracoes realizadas
        *cont = passos;

        // Supomos ter atingido a idempotencia
        idempotencia = 1;

        // Verificando se a suposicao eh mesmo correta
        for(i = 0; i < nl * nc; i++)
        {
            if(*(imagemSaida + i) != *(imagemAnterior + i))
            {
                // Suposicao errada: nao chegamos na idempotencia entao
                // realizar passo adicional
                idempotencia = 0;
                copiarImagem(imagemSaida, imagemAnterior, nl, nc);
            }
        }
    }

    desaloca_memo(&imagemErodida);
    desaloca_memo(&imagemDilatada);
    desaloca_memo(&imagemAnterior);
}

// Copia uma imagem pixel a pixel
void copiarImagem(int *fonte, int *destino,  int nl, int nc)
{
    int i;

    for(i = 0; i < nl * nc; i++)
        *(destino + i) = *(fonte + i);
}

// Zera todos os pixels de uma imagem
void zerarImagem(int *imagem, int nl, int nc)
{
    int i, j;

    // Zerando imagem
    for(i = 0; i < nl; i++)
    {
        for(j = 0; j < nl; j++)
            *( imagem + i * nc + j) = 0;
    }
}

// B = elemento estruturante
// Faz a erosao/dilatacao no espaco multiescala
// Retorna 1 se atingiu a idempotencia
// iter pode servir para indicar o numero de iteracoes
int erosaoDilatacaoMMDE(int *imagem, int *imagemSaida, int nl, int nc, int B[TAM_JANELA][TAM_JANELA],
                         int op, int iter)
{
    int c, l, nj, i, j, k, i1, i2;
    int janela[TAM_JANELA][TAM_JANELA];
    int x;
    int elementoCentral = 0;
    int contador = 0;
    int resultadoParcial;
    int diferenca, soma;

    // Reflete qual o elemento da janela passada deve ser alterado
    int posicao = (TAM_JANELA * TAM_JANELA)/ 2 + 1;

    nj = 0;
    x = 0;

    zerarImagem(imagemSaida, nl, nc);

    for(x = 0; x < nl - TAM_JANELA + 1; x++)
    {
        for(nj = 0; nj < nc - TAM_JANELA + 1; nj++)
        {
            k = nj + x * nc;
            l = 0;
            contador = 0;

            for(i = 0; i < TAM_JANELA; i++)
            {
                c = 0;
                for(j = k; j < (TAM_JANELA + k); j++)
                {
                    contador++;
                    janela[l][c] = *(imagem + j);
                    c++;

                    if(contador == posicao)
                        elementoCentral = j;

                }
                k = (k + nc);
                l++;
            }

            if(op == 0)
            {
                resultadoParcial = INFINITO;

                for(i1 = 0; i1 < TAM_JANELA; i1++)
                {
                    for(i2 = 0; i2 < TAM_JANELA; i2++)
                    {
                        // Esta considerando o elemento central na comparacao
                        if(janela[i1][i2] - B[i1][i2] <= resultadoParcial)
                        {
                            // Fazendo a verificacao para limitar a erosao
                            // entre [0 - 255]
                            // Estas definicoes estao no arquivo PDF capitulo 5
                            // que te mandei "Bounded gray-scale morphology"
                            diferenca = janela[i1][i2] - B[i1][i2];

                            if(janela[i1][i2] < 255 && diferenca <= 0)
                                resultadoParcial = 0;
                            else if(janela[i1][i2] < 255 && (0 <= diferenca) && (diferenca <= 255))
                                resultadoParcial = diferenca;
                            else if(janela[i1][i2] < 255 && diferenca > 255)
                                resultadoParcial = 255;
                            else
                                resultadoParcial = 255;
                        }
                    }
                }
            }
            else
            {
                resultadoParcial = 0;

                for(i1 = 0; i1 < TAM_JANELA; i1++)
                {
                    for(i2 = 0; i2 < TAM_JANELA; i2++)
                    {
                        if(janela[i1][i2] + B[i1][i2] >= resultadoParcial)
                        {
                            soma = janela[i1][i2] + B[i1][i2];

                            if(janela[i1][i2] == 0)
                                resultadoParcial = 0;
                            else if(janela[i1][i2] > 0 && soma <= 0)
                                resultadoParcial = 0;
                            else if(janela[i1][i2] > 0 && (0 <= janela[i1][i2]) && (soma <= 255))
                                resultadoParcial = soma;
                            else
                                resultadoParcial = 255;
                        }
                    }
                }
            }

           *(imagemSaida + elementoCentral) = resultadoParcial;
        }
    }

    // Verificando a idempotencia
    i = 0;
    while(*(imagemSaida + i) == *(imagem + i) && i < nl * nc)
        i++;

    if(i >= (nl * nc))
        return 1;
    else
        return 0;
}

// Ler de um arquivo uma sequencia de escalas de modo a conseguir
// aplicar o filtro alternado sequencial para cada escala
void configurarParametros(char *nomeArquivo,  float *listaEscalas)
{
    FILE *fp;
    char str[40];
    int i = 0;

    /* Arquivo ASCII, para leitura */
    fp = fopen(nomeArquivo,"r");

    if(!fp)
    {
        printf("Erro na abertura do arquivo");
        exit(0);
    }

    while(fgets(str, 40, fp) != NULL && i < NRO_ESCALAS)
    {
        *(listaEscalas + i) = atof(str);
        i++;
    }
}
